import numpy as np
import cluster_tree
import sys
from counting_matrix import *



class ClusterAlg(object):
    """
    A abstract clustering algorithm. All clustering algorithms should inherit from
    this. The interface exposed is that each call to ClusterAlg.cluster should
    recieve a similarity matrix M and terminate with a ClusterTree object in the
    result instance variable. Each ClusterAlg should have a name that can be used to
    identify it as well as a shortname that can be used for logging and writing
    files.
    """
    def __init__(self):
        """
        Construct a ClusterAlg.
        """
        self.result = None
        self.name = "cluster_alg"

    def cluster(self, M, k=2, params={}):
        """
        Default clustering algorithm, just splits the number of objects
        evenly. Note that after completion of this algorithm, self.result is
        assigned to a ClusterTree object.
        """
        if type(M) == CountingMatrix:
            n = M.data.shape[0]
        else:
            n = M.shape[0]
        if 'stopping_cond' in params.keys() and params['stopping_cond'](n):
            self.result = self.make_ctree(n, [set(range(n))])
            return
        sets = [set([i for i in range(n*j/k, n*(j+1)/k)]) for j in range(k)]
        self.result = self.make_ctree(n, sets)

    def preprocess(self, M):
        """
        Essentially a macro for handling CountingMatrix and np.matrix in the same way.
        """
        if (type(M) == CountingMatrix):
            return M.all()
        return M

    def make_ctree(self, n, sets):
        """
        Construct a ClusterTree given a set of flat clusters. This is a very
        handy subroutine for all of the flat clustering algorithms, as well as
        for when hierarchical algorithms hit the bottom of the recursion.
        """
        if len(sets) == 1:
            return cluster_tree.ClusterTree([[0 for i in range(len(sets[0]))]])
        lst = []
        clists = []
        for x in range(n):
            lst.extend([i for i in range(len(sets)) if x in sets[i]])
        for j in range(len(sets)):
            clists.append([[0 for x in sets[j]]])
        lst = [lst]
        lst.extend(clists)
        return cluster_tree.ClusterTree(lst)


class HClusterAlg(object):
    """
    This is a hierarchical clustering algorithm. It needs as input a ClusterAlg
    object that it can use to perform splits at each level.
    """
    def __init__(self, calg):
        """
        Initialize a hierarchical clustering algorithm with a flat one that it
        can use a subroutine.
        """
        self.alg = calg
        self.result = None
        self.name = "hcluster_alg"

    def cluster(self, M, k=2, params={}):
        """
        Call self.alg on all of the subclusters. Recursively performs top-down
        splits. Again note that the result is populated with a ClusterTree Object
        """
        def cluster_rec(M, curr, k, params):
            ## Check if we've reached the bottom of the recursion
            if len(curr) <= 1 or ('stopping_cond' in params.keys() and params['stopping_cond'](len(curr))):
                return [[0 for i in curr]]

            ## Otherwise we have to cluster the submatrix
            if type(M) == CountingMatrix:
                self.alg.cluster(M.submatrix(curr), k, params)
            else:
                self.alg.cluster(M[np.ix_(curr, curr)], k, params)

            ## Another case of the end of the recursion, we just found one cluster.
            flat_ct = self.alg.result
            if len(flat_ct.sets) == 1:
                ## this is the end of the recursion, just return 
                return flat_ct.lists
            ## BOOM list comprehensions for the win. Just postprocessing output.
            tmpout = [[curr[i] for i in se] for se in flat_ct.sets[1:]]
            recs = [cluster_rec(M, [curr[i] for i in se], k, params) for se in flat_ct.sets[1:]]
            out = [[i for i in range(len(tmpout)) if x in tmpout[i]][0] for x in curr]
            out = [out]
            out.extend(recs)
            return out
        n = M.shape[0]
        out = cluster_rec(M, range(n), k, params)
        self.result = cluster_tree.ClusterTree(out)
