import numpy as np
import scipy.sparse
import matplotlib.pyplot as plt

class Matrix(object):
    """
    A simple matrix object. 
    """
    def __init__(self, arr):
        """
        Constructor -- takes in an np.matrix.
        """
        self.data = arr
        self.shape = arr.shape

    def get(self, i,j):
        """
        Fetch the entry at M[i,j]
        """
        return self.data[i,j]

    def submatrix(self, indices):
        """
        Fetch a square submatrix of this matrix, represented by the rows and
        columns in indices.
        """
        return Matrix(self.data[np.ix_(indices, indices)])

    def row(self, i):
        """
        Get one specific row (actually this is a column"
        """
        return self.data[:,i]

    def all(self):
        """
        Return all of the items of this matrix.
        """
        return self.data

    def markasused(self, xs, ys):
        pass

    def show(self, file=None):
        """
        Render the heat map for this matrix and save it to the file.
        """
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        ax1.imshow(self.data)
        if file != None:
            plt.savefig(file)
        fig.clf()
        plt.close(fig)

class CountingMatrix(Matrix):
    """
    A matrix that also maintains information about which entries have been
    observed by an algorithm.

    Since we can do subindexing we have to keep track of where this matrix
    belongs in a potential parent. In this case we do not have a parent, but
    once we start fetching submatrices, we will have to maintain that
    information.
    """
    def __init__(self, arr, type=scipy.sparse.dok_matrix):
        """
        Construtor, initialize the used matrix to be all zeros. Set parent to be None
        """
        self.data = arr
        if type == np.matrix:
            self.used = np.matrix(np.zeros(arr.shape))
        else:
            self.used = type(arr.shape)
        self.shape = self.data.shape
        self.parent = None
        self.mapping = None

    def get(self, i,j):
        """
        Fetch an element and mark it as used. Mark it as used in all parents of this matrix if they exist.
        """
        self.markasused([i], [j])
        return self.data[i,j]

    def numUsed(self):
        """
        Fetch the number of used elements, since all matrices we deal with are
        symmetric, we only count entries in the upper triangle, not including
        the main diagonal.
        """
        return scipy.sparse.triu(self.used).getnnz()

    def submatrix(self, indices):
        """
        Return a new CountingMatrix that is a submatrix of this one, specified
        by the rows/columns indices. That matrix has this one as a parent, so
        that observations there are reflected here.
        """
        Cnew = CountingMatrix(self.data[np.ix_(indices, indices)], type=type(self.used))
        Cnew.parent = self
        Cnew.mapping = indices
        return Cnew

    def row(self, i):
        """
        Fetch a row of this matrix and mark all entries as used.
        """
        self.markasused(i, range(self.data.shape[0]))
        return self.data[:,i]

    def all(self):
        """
        Fetch the entire matrix, marking it all as used. 
        """
        self.markasused(range(self.data.shape[0]), range(self.data.shape[1]))
        return self.data

    def markasused(self, xs, ys):
        """
        Mark a set of entries as used, specifically the submatrix
        M[np.ix_(xs,ys)] is marked in this matrix and all parents.
        """
        if type(self.used) == np.matrix:
            self.used[np.ix_(xs, ys)] = np.ones((len(xs), len(ys)))
            self.used[np.ix_(ys, xs)] = np.ones((len(ys), len(xs)))
        else:
            self.used[xs,ys] = np.ones((len(xs), len(ys)))
            self.used[ys,xs] = np.ones((len(ys), len(xs)))
        if self.parent != None:
            mxs = [self.mapping[x] for x in xs]
            mys = [self.mapping[y] for y in ys]
            self.parent.markasused(mxs, mys)

    def getall(self, W):
        """
        Given a bernoulli matrix, return a matrix with my entries where W is 1
        and zeros everywhere else
        """
        out = np.matrix(np.array(self.data)*np.array(W))
        for i in range(W.shape[0]):
            for j in range(W.shape[1]):
                if W[i,j] == 1:
                    self.markasused([i],[j])
        return out

    def show(self, file=None):
        """
        Render a heat map of this matrix, restricted only to the observed
        entries.
        """
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        ax1.imshow(np.array(self.data)*np.array(self.used))
        if file != None:
            plt.savefig(file, dpi=500, format="eps")
        fig.clf()
        plt.close(fig)
