import numpy as np
import matrices, cluster_alg, time, util, alg_names, settings
import sys, os, copy, traceback
import hierarchical_algorithms, subsample_objects
from counting_matrix import *

class Experiment(object):
    """
    Given a clustering algorithm, the Experiment class provides routines to run
    several experiments for a specific algorithm.
    """
    def __init__(self, alg, file=None):
        """
        Initialize an Experiment object. 
        """
        self.alg = alg
        self.tc_n_scores = {}
        self.tc_n_mcs = {}
        self.tc_s_scores = {}
        self.tc_s_logexp_scores = {}
        self.runtime_scores = {}
        self.file = file
        if self.file == None:
            self.file = sys.stdout

    def _threshold_curve_sub(self, n, sigmas, reps, params={}):
        """
        Private method. Called by threshold_curves
        """
        (M, ct, name) = matrices.make_ultrametric(n, 0, 0.5)
        scores = []
        truth_alg = cluster_alg.ClusterAlg()
        truth_alg.cluster(M,k=2)
        mc_means = []
        mc_vars = []
        for i in sigmas:
            correct = 0
            tmp_mcs = []
            for k in range(reps):
                A = M.data + matrices.make_gaussian_noise(n, i)
                self.alg.cluster(A, k=2, params=params)
                mc = self.alg.result.misclustering(truth_alg.result)
                if mc[0] == 0:
                    correct += 1
                    tmp_mcs.append(mc[0])
                mc_means.append(np.mean(tmp_mcs))
                mc_vars.append(np.var(tmp_mcs))
            scores.append(float(correct)/reps)
        return (scores, mc_means, mc_vars)

    def threshold_curves(self, 
                         ns=(128, 256, 512, 1024), 
                         sigmas=np.arange(0.1, 2.1, 0.1), 
                         matrixfunc = lambda x: matrices.make_ultrametric(x, 0, 0.5),
                         params={}):
        """
        Using ultrametric matrices, compute threshold curves using probability of
        correctness as a metric for success. Threshold curves plot probability of
        recovering the first split as a function of the noise variance for fixed
        problem size n.

        Write output to settings.OUTDIR
        """
        (M, ct, mname) = matrixfunc(ns[0])
        if self.file != None:
            self.file.write("Starting Threshold Curve Experiments for ALG=%s, matrix=%s\n" % (self.alg.name, mname))
            self.file.flush()
            os.fsync(self.file)

        reps = 20
        for n in ns:
            if self.file != None:
                self.file.write("Starting n=%d\n" % (n))
                self.file.flush()
                os.fsync(self.file)
            (tmpscores, mc_means, mc_vars) = self._threshold_curve_sub(n, sigmas, reps)
            self.tc_n_scores[n] = tmpscores
            print "done with %d" % (n)
                
        if self.file != None:
            self.file.write("Writing to out files\n")

        f = open(settings.OUTDIR+"%s_tc_n_%s.out" % (self.alg.name, mname), "w")
        s1 = "sigma " + " ".join([str(sigma) for sigma in sigmas])
        f.write("%s\n" % (s1))
        for n in ns:
            f.write("%s\n" % (str(n) + " " + " ".join([str(x) for x in self.tc_n_scores[n]])))
        f.close()
        if self.file != None:
            self.file.write("Done\n")
            self.file.flush()
            os.fsync(self.file)
        
    def threshold_curves_s(self,
                           n=2048, 
                           s=(10,20,30), 
                           sigmas = np.arange(0.1, 1.01, 0.05),
                           matrixfunc = lambda x: matrices.make_ultrametric(x, 0, 0.5),
                           params={},
                           logexp=False):
        """
        Repeat the threshold_curves experiment for different values of s, which
        controls the number of similarities used by the algorithm. Here n is
        fixed. The parameter logexp determines how s changes the number of
        similarities, if False, s is just a multiplicative factor, otherwise it
        appears in the exponent.
        
        Writes output to settings.OUTDIR
        """
        reps = 20
        scores = {}
        for ss in s:
            if logexp:
                params['s'] = lambda x: np.ceil(np.log(x)**ss)
            else:
                params['s'] = lambda x: np.ceil(ss*np.log(x))
            (tmpscores, mc_means, mc_vars) = self._threshold_curve_sub(n, sigmas, reps, params=params)
            scores[ss] = tmpscores
        (M, ct, mname) = matrixfunc(n)
        if logexp == True:
            f = open(settings.OUTDIR+"%s_tc_s_%s_logexp.out" % (self.alg.name, mname), "w")
            self.tc_s_logexp_scores = scores
        else:
            f = open(settings.OUTDIR+"%s_tc_s_%s.out" % (self.alg.name, mname), "w")
            self.tc_s_scores = scores
        s1 = "s " + " ".join([str(sigma) for sigma in sigmas])
        f.write("%s\n" % (s1))
        for ss in s:
            f.write("%s\n" % (str(n) + " " + " ".join([str(x) for x in scores[ss]])))
        f.close()

    def runtime(self, ns = [2**i for i in range(5, 12)], params={}):
        """
        Measure the running time of an algorithm as a function of n. This does not
        check correctness or make an attempt to add noise. 
        
        Writes output to settings.OUTDIR
        """
        reps = 25
        ms = []
        vs = []
        if self.file != None:
            self.file.write("Starting Runtime Experiments for ALG=%s\n" % (self.alg.name))
            self.file.flush()
            os.fsync(self.file)
        for n in ns:
            if self.file != None:
                self.file.write("Starting n=%d\n" % (n))
                self.file.flush()
                os.fsync(self.file)
            scores = []
            M = matrices.make_ultrametric(n, 0, 0.5)[0].data
            A = M + matrices.make_gaussian_noise(n, 0.001)
            for k in range(reps):
                start = time.time()
                self.alg.cluster(M, params=copy.deepcopy(params))
                stop = time.time()
                scores.append(stop-start)
            ms.append(np.mean(scores))
            vs.append(np.var(scores))

        if self.file != None:
            self.file.write("Writing to out files\n")
        f = open(settings.OUTDIR+"%s_runtime.out" % self.alg.name, "w")
        s1 = "n " + " ".join([str(n) for n in ns])
        f.write("%s\n" % (s1))
        f.write("means %s\n" % (" ".join([str(x) for x in ms])))
        f.write("vars %s\n" % (" ".join([str(x) for x in vs])))
        f.close()

        self.runtime_scores = (ms, vs)
        if self.file != None:
            self.file.write("Done\n")
            self.file.flush()
            os.fsync(self.file)
    
    def probes(self, ns = [2**i for i in range(5,12)], params={}):
        """
        Measure the sample complexity of an algorithm as a function of n. This does
        not check correctness or make an attempt to add noise.
        
        Writes output to settings.OUTDIR
        """
        reps = 10
        ms = []
        vs = []
        if self.file != None:
            self.file.write("Starting Probing Experiments for ALG=%s\n" % (self.alg.name))
            self.file.flush()
            os.fsync(self.file)
        for n in ns:
            if self.file != None:
                self.file.write("Starting n=%d\n" % (n))
                self.file.flush()
                os.fsync(self.file)
            scores = []
            for k in range(reps):
                M = matrices.make_ultrametric(n, 0, 0.5)[0]
                self.alg.cluster(M, params=copy.deepcopy(params))
                scores.append(M.numUsed())
            ms.append(np.mean(scores))
            vs.append(np.std(scores))

        if self.file != None:
            self.file.write("Writing to out files\n")
        f = open(settings.OUTDIR+"%s_probes.out" % self.alg.name, "w")
        s1 = "n " + " ".join([str(n) for n in ns])
        f.write("%s\n" % (s1))
        f.write("means %s\n" % (" ".join([str(x) for x in ms])))
        f.write("vars %s\n" % (" ".join([str(x) for x in vs])))
        f.close()
        self.probes_scores = (ms, vs)
        if self.file != None:
            self.file.write("Done\n")
            self.file.flush()
            os.fsync(self.file)

    def misclustering(self, n=1024,
                      sigmas=np.arange(0.1, 1, 0.05),
                      matrixfunc = lambda x: matrices.make_ultrametric(x, 0, 0.5),
                      params={}):
        """
        Measure the outlier fraction of an algorithm as a function of sigma for
        fixed n
        
        Writes output to settings.OUTDIR
        """
        reps = 25
        mc_ms = []
        mc_vs = []
        of_ms = []
        of_vs = []
        (M, ct, mname) = matrixfunc(n)
        if self.file != None:
            self.file.write("Starting Misclustering Experiments for ALG=%s, n=%d, matrix=%s\n" % (self.alg.name, n, mname))
            self.file.flush()
            os.fsync(self.file)
        for s in sigmas:
            if self.file != None:
                self.file.write("Starting sigma=%f\n" % (s))
                self.file.flush()
                os.fsync(self.file)
            mc_scores = []
            of_scores = []
            for k in range(reps):
                A = M.data+matrices.make_gaussian_noise(n, s)
                self.alg.cluster(A, params=copy.deepcopy(params))
#                 mc_scores.append(self.alg.result.misclustering(ct)[0])
                of_scores.append(self.alg.result.outlierfraction(ct, sampled=True))
            mc_ms.append(np.mean(mc_scores))
            mc_vs.append(np.var(mc_scores))
            of_ms.append(np.mean(of_scores))
            of_vs.append(np.var(of_scores))
        if self.file != None:
            self.file.write("Writing to out files\n")
#         f = open(settings.OUTDIR+"%s_misclustering_n=%d_%s.out" % (self.alg.name, n, mname), "w")
#         s1 = "sigma " + " ".join([str(sigma) for sigma in sigmas])
#         f.write("%s\n" % (s1))
#         f.write("means %s\n" % (" ".join([str(x) for x in mc_ms])))
#         f.write("vars %s\n" % (" ".join([str(x) for x in mc_vs])))
#         f.close()

        f = open(settings.OUTDIR+"%s_outlier_n=%d_%s.out" % (self.alg.name, n, mname), "w")
        s1 = "sigma " + " ".join([str(sigma) for sigma in sigmas])
        f.write("%s\n" % (s1))
        f.write("means %s\n" % (" ".join([str(x) for x in of_ms])))
        f.write("vars %s\n" % (" ".join([str(x) for x in of_vs])))
        f.close()

        self.mc_scores = (mc_ms, mc_vs)
        self.of_scores = (of_ms, of_vs)
        if self.file != None:
            self.file.write("Done\n")
            self.file.flush()
            os.fsync(self.file)


def allruntimes(ns = [2**i for i in range(5,13)], params={}):
    """
    Generate the raw data for all of the runtime curves.
    """
    ## first delete all the related files.
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            os.remove(settings.OUTDIR+"%s_runtime.out" %(alg().name))
        except OSError:
            continue
    f = open(settings.LOGDIR+"runtime.log", "w")
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            E = Experiment(alg(), file=f)
            E.runtime(ns, params)
        except (Error, RuntimeException) as inst:
            f.write("Caught %s\n" % str(type(inst)))
            traceback.print_exc(file)
            continue

def allprobes(ns = [2**i for i in range(5,13)], params={}):
    """
    Generate the raw data for all of the probing curves.
    """
    ## first delete all the related files.
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            os.remove(settings.OUTDIR+"%s_probes.out" %(alg().name))
        except OSError:
            continue
    f = open(settings.LOGDIR+"probes.log", "w")
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            E = Experiment(alg(), file=f)
            E.probes(ns, params=copy.deepcopy(params))
        except (Exception, RuntimeError) as inst:
            f.write("Caught %s\n" % str(type(inst)))
            traceback.print_exc(f)
            continue

def allmisclusterings(n = 1024, 
                      sigmas = np.arange(0.05, 1.05, 0.05), 
                      matrixfunc = lambda x: matrices.make_ultrametric(x, 0, 0.5), 
                      params = {}):
    """
    Generate the raw data for all of the probing curves.
    """
    ## first delete all the related files.
    (M, ct, mname) = matrixfunc(n)
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            os.remove(settings.OUTDIR+"%s_misclustering_n=%d_%s.out" %(alg().name, n, mname))
        except OSError:
            pass
        try:
            os.remove(settings.OUTDIR+"%s_outlier_n=%d_%s.out" %(alg().name, n, mname))
        except OSError:
            pass
    f = open(settings.LOGDIR+"misclustering_n=%d_%s.log" % (n, mname), "w")
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            E = Experiment(alg(), file=f)
            E.misclustering(n=n, 
                            sigmas=sigmas, 
                            matrixfunc=matrixfunc, 
                            params=copy.deepcopy(params))
        except (Exception, RuntimeError) as inst:
            f.write("Caught %s\n" % str(type(inst)))
            traceback.print_exc(f)
            continue

def allthresholdcurves(ns = (32, 64, 128), 
                       matrixfunc = lambda x: matrices.make_ultrametric(x, 0, 0.5),
                       params = {}):
    """
    Generate raw data for threshold curves for all of the algorithms.
    """
    (M, ct, mname) = matrixfunc(ns[0])
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            os.remove(settings.OUTDIR+"%s_tc_n_%s.out" %(alg().name, mname))
        except OSError:
            pass
    f = open(settings.LOGDIR+"tc_n_%s_large.log" % (mname), "w")
    for alg in alg_names.HIERARCHICAL_ALGORITHMS:
        try:
            E = Experiment(alg(), file=f)
            E.threshold_curves(ns=ns, 
                               matrixfunc=matrixfunc, 
                               params=copy.deepcopy(params))
        except (Exception, RuntimeError) as inst:
            f.write("Caught %s\n" % str(type(inst)))
            traceback.print_exc(f)
            continue
    
