import numpy as np
import flat_algorithms, cluster_tree, subsample_objects
from cluster_alg import *
import sys, copy
from counting_matrix import *


"""
This module contains algorithms for doing hierarchical clustering.
"""
class HierarchicalSpectral(HClusterAlg):
    """
    All we have to do here is provide the algorithm for single splits. In this
    case it is just SpectralMethod.
    """
    def __init__(self):
        super(HierarchicalSpectral, self).__init__(flat_algorithms.SpectralMethod())
        self.name = "hierarchical_spectral"
        self.shortname = "hs"

class SingleLinkage(HClusterAlg):
    """
    Standard Single Linkage Clustering, but modify the output so that it is in
    the same format as all of the other algorithms.
    """
    def __init__(self, method="single"):
        super(SingleLinkage, self).__init__(None)
        self.calg = None
        self.result = None
        self.method = method
        self.name = self.method +"_linkage"
        self.shortname = self.method[0]+"l"


    def cluster(self, M, k=2, s=lambda x: x, params={}):
        if (type(M) == CountingMatrix):
            M = M.all()
        A = M
#         A = np.minimum(A, 0.99999999*np.ones((M.shape[0], M.shape[0])))
#         A = np.maximum(A, np.zeros((M.shape[0], M.shape[0])))
#         X = -np.log(A + np.diag([1 - A[i,i] for i in range(A.shape[0])]))
        A = A - np.min(A) + 0.00001
        A = A/np.max(A)
        X = -np.log(A + np.diag([1 - A[i,i] for i in range(A.shape[0])]))
        vec = scipy.spatial.distance.squareform(X)
        out = scipy.cluster.hierarchy.linkage(vec, method=self.method)
        r = scipy.cluster.hierarchy.to_tree(out)
        def cl_below2(Tn):
            if Tn.is_leaf():
                return ([[0]], [Tn.get_id()])
            else:
                (c1, id1) = cl_below2(Tn.get_left())
                (c2, id2) = cl_below2(Tn.get_right())
                mapping_to_return = copy.copy(id1)
                mapping_to_return.extend(id2)
                curr_clus = [0 if x in id1 else 1 for x in mapping_to_return]
                return ([curr_clus, c1, c2], mapping_to_return)
        (cls, blah) = cl_below2(r)
        self.result = cluster_tree.ClusterTree(cls)


class HierarchicalKMeans(HClusterAlg):
    """
    Recursively subpartitions by running k-means with k=2 on each subcluster.
    """
    def __init__(self):
        super(HierarchicalKMeans, self).__init__(flat_algorithms.KMeans())
        self.name="hierarchical_kmeans"
        self.shortname = "hkm"

    def cluster(self, M, k=2, params={}):
        params['active'] = False
        super(HierarchicalKMeans, self).cluster(M, k=k, params=params)

class HierarchicalActiveKMeans(HClusterAlg):
    """
    Active version of hierarchical K-Means. At each split, subsample a set of
    coordinates and use only those coordinates for computing distances in the
    k-means algorithm. Defaults to log n coordinates to subsample (out of a
    possible n)
    """
    def __init__(self):
        super(HierarchicalActiveKMeans, self).__init__(flat_algorithms.KMeans())
        self.name="hierarchical_active_kmeans"
        self.shortname = "hakm"

    def cluster(self, M, k=2, params={}):
        params['active'] = True
        if 's' not in params.keys():
            params['s'] = lambda x: np.ceil(np.log2(x))
        if 'stopping_cond' not in params.keys():
            params['stopping_cond'] = params['s']
        params['drop'] = False
        params['make_new'] = False
        n = M.shape[0]
        ssize = params['s'](n)
        params['s'] = lambda x: ssize
        scond = params['stopping_cond'](n)
        params['stopping_cond'] = lambda x: x <= scond
        super(HierarchicalActiveKMeans, self).cluster(M, k=k, params=params)

class HierarchicalSOWithHeuristics(HClusterAlg):
    """
    Hierarchical Active Spectral Clustering algorithm with various
    heuristics. Subsamples a log n x log n square of the matrix, run spectral
    clustering on that and then performs a simple post-processing step where it
    places every remaining object in the cluster with highest similarity. The heuristics are:

    1. drop low degree nodes in the spectral clustering phase.
    2. if average similarity to existing clusters is too low, then create a new
    cluster.
    """
    def __init__(self):
        super(HierarchicalSOWithHeuristics, self).__init__(subsample_objects.SOWithHeuristics())
        self.name = "hierarchical_so_heuristic"
        self.shortname = "hash"

    def cluster(self, M, k=2, params={}):
        ## Make sure that the parameters are set correctly. 
        params['active'] = True
        if 's' not in params.keys():
            params['s'] = lambda x: np.ceil(np.log2(x))
        if 'stopping_cond' not in params.keys():
            params['stopping_cond'] = lambda x: np.ceil(np.log2(x))

        params['make_new'] = True
        params['drop'] = True
        n = M.shape[0]
        ## place the level of activeness into the algorithm, by providing it in
        ## the params.
        ssize = params['s'](n)
        params['s'] = lambda x: ssize
        scond = params['stopping_cond'](n)
        params['stopping_cond'] = lambda x: x <= scond
        super(HierarchicalSOWithHeuristics, self).cluster(M, k=k, params=params)

class HierarchicalSO(HClusterAlg):
    """
    Hierarchical Active Spectral Clustering algorithms. Subsamples a log n x log
    n square of the matrix, run spectral clustering on that and then performs a
    simple post-processing step where it places every remaining object in the
    cluster with highest similarity.
    """
    def __init__(self):
        super(HierarchicalSO, self).__init__(subsample_objects.SOWithHeuristics())
        self.name = "hierarchical_so"
        self.shortname = "has"

    def cluster(self, M, k=2, params={}):
        params['active'] = True
        if 's' not in params.keys():
            params['s'] = lambda x: np.ceil(np.log2(x))
        if 'stopping_cond' not in params.keys():
            params['stopping_cond'] = params['s']
        params['drop'] = False
        params['make_new'] = False
        n = M.shape[0]
        ssize = params['s'](n)
        params['s'] = lambda x: ssize
        scond = params['stopping_cond'](n)
        params['stopping_cond'] = lambda x: x <= scond
        super(HierarchicalSO, self).cluster(M, k=k, params=params)
