import numpy as np
import sys
import util, cluster_tree
from counting_matrix import *


"""
This module contains a lot of the input matrices we consider.
"""
def make_ultrametric(n, old, gap):
    """
    This is a balanced binary ultrametric. Gaps are decreasing as they should.
    """
    M = np.matrix(np.zeros((n,n)))
    if n == 1:
        M[0,0] = old+gap;
        return (CountingMatrix(M), cluster_tree.ClusterTree([[0]]), "um")
    M[np.ix_(range(n/2, n), range(n/2))] = old+gap
    M[np.ix_(range(n/2), range(n/2, n))] = old+gap
    old = old+gap
    (Msub, ct1, blah) = make_ultrametric(n/2, old, gap/2)
    M[np.ix_(range(n/2), range(n/2))] = Msub.data
    (Msub, ct2, blah) = make_ultrametric(n/2, old, gap/2)
    M[np.ix_(range(n/2, n), range(n/2, n))] = Msub.data
    lst = [0 for i in range(n/2)]
    lst.extend([1 for i in range(n/2, n)])
    ct = cluster_tree.ClusterTree([lst, ct1.lists, ct2.lists])
    return (CountingMatrix(M, type=np.matrix), ct, "um")

def make_hierarchical_block(n):
    """
    This is the hierarchical block matrix with non-constant off-diagonals. Again
    it is perfectly balanced.
    """
    def hb_recurse(n, low, high, gap):
        M = np.zeros((n,n))
        if n == 1:
            M[0,0] = np.random.uniform(low, high, [1, 1])
            return (M, cluster_tree.ClusterTree([[0]]))
        M[np.ix_(range(n/2,n), range(n/2))] = np.random.uniform(low, high, [len(range(n/2,n)), n/2])
        M = M+M.T
        r = 1.0/2*(high-low)
        low = high+gap
        high = high+gap+r
        gap = gap/2
        (Msub, ct1) = hb_recurse(n/2, low, high, gap)
        M[np.ix_(range(n/2), range(n/2))] = Msub
        (Msub, ct2) = hb_recurse(len(range(n/2,n)), low, high, gap)
        M[np.ix_(range(n/2,n), range(n/2,n))] = Msub
        lst = [0 for i in range(n/2)]
        lst.extend([1 for i in range(n/2, n)])
        ct = cluster_tree.ClusterTree([lst, ct1.lists, ct2.lists])
        return (M, ct)
    (M, ct) = hb_recurse(n, 0.25, 0.5, 0.25)
    return (CountingMatrix(M, type=np.matrix), ct, "hb")

def ultrametric_eig(n, old, gap):
    """
    This is the sum of the eigenvalues for an ultrametric with the specific
    parameters.
    """
    s = [old + gap*(2**i-1)/2**(i-1) for i in range(1, int(np.log2(n)+2))]
    return sum([np.ceil(float(n)/2**i) * s[i-1] for i in range(1, int(np.log2(n))+2)])

def make_kway(n, k, beta0, gap, r):
    """
    Construct a k-way matrix. The off-diagonal is constant at beta0, The
    diagonals are randomly generated between beta0+gap and beta0+gap+r.
    """
    if np.ceil(float(n)/k) != n/k:
        raise Exception("n must be a multiple of k")

    M = beta0*np.matrix(np.ones((n,n)))
    ct = []
    lsts = []
    for i in range(k):
        X = np.matrix(np.random.uniform(beta0+gap, beta0+gap+r, (n/k, n/k)))
        M[np.ix_(range(n/k*(i), n/k*(i+1)),range(n/k*i, n/k*(i+1)))] = 1.0/2*(X + X.T)
        ct.extend([i for x in range(n/k*(i), n/k*(i+1))])
        lsts.append([[0 for x in range(n/k*(i), n/k*(i+1))]])
    ct = [ct]
    ct.extend(lsts)
    return (CountingMatrix(np.matrix(M), type=np.matrix), cluster_tree.ClusterTree(ct), "k")


def make_stochastic_block(n, k, p, q):
    """
    Make a k-way stochastic block model graph. This is a binary graph where
    on-diagonal edges have probability p of occuring and off-diagonal edges have
    probability q.
    """
    if np.ceil(float(n)/k) != n/k:
        raise Exception("n must be a multiple of k")
    M = np.matrix(np.random.binomial(1, q, (n,n)))
    M = np.triu(M,k=1)
    M = M+M.T
    ct = []
    lsts = []
    for i in range(k):
        X = np.matrix(np.random.binomial(1, p, (n/k, n/k)))
        X = np.triu(X, k=1) 
        X = X+X.T + np.matrix(np.diagflat(np.array([1 for x in range(n/k)])))
        M[np.ix_(range(n/k*(i), n/k*(i+1)),range(n/k*i, n/k*(i+1)))] = X
        ct.extend([i for x in range(n/k*(i), n/k*(i+1))])
        lsts.append([[0 for x in range(n/k*(i), n/k*(i+1))]])
    ct = [ct]
    ct.extend(lsts)
    return (CountingMatrix(np.matrix(M), type=np.matrix), cluster_tree.ClusterTree(ct), "sb")


def make_hierarchical_planted_partition(n, p):
    """
    Construct a hierarchical planted partition. Here edges in the off-diagonal
    at level i have probability p[i] of occuring. p can be a list of
    probabilities or it is automatically increased as we move down the hierarchy
    (so the clustering is well formed). This generates a balanced binary
    hierarchy.
    """
    if n == 1:
        return (CountingMatrix(np.matrix(np.ones((1,1)))), cluster_tree.ClusterTree([[0]]), "hpp")
    M = np.matrix(np.zeros((n,n)))
    if type(p) == list:
        if len(p) == 1:
            p = p[0]
        else:
            M[0:n/2, n/2:n] = np.random.binomial(1, p[0], (len(range(0, n/2)), len(range(n/2, n))))
            p = p[1:len(p)]
            M[n/2:n, 0:n/2] = M[0:n/2, n/2:n].T
            (Msub, ct1, blah) = make_hierarchical_planted_partition(len(range(0,n/2)), p)
            M[0:n/2, 0:n/2] = Msub.data
            (Msub, ct2, blah) = make_hierarchical_planted_partition(len(range(n/2,n)), p)
            M[n/2:n, n/2:n] = Msub.data
    if type(p) != list:
            M[0:n/2, n/2:n] = np.random.binomial(1, p, (len(range(0, n/2)), len(range(n/2, n))))
            M[n/2:n, 0:n/2] = M[0:n/2, n/2:n].T
            (Msub, ct1, blah) = make_hierarchical_planted_partition(len(range(0,n/2)), p+float(1-p)/2)
            M[0:n/2, 0:n/2] = Msub.data
            (Msub, ct2, blah) = make_hierarchical_planted_partition(len(range(n/2, n)), p+float(1-p)/2)
            M[n/2:n, n/2:n] = Msub.data
    lst = [0 for i in range(n/2)]
    lst.extend([1 for i in range(n/2, n)])
    ct = cluster_tree.ClusterTree([lst, ct1.lists, ct2.lists])
    return (CountingMatrix(M, type=np.matrix), ct, "hpp")

def make_arbitrary_hierarchical_block(n,p, alpha=0.5, r=True):
    """
    Makes a hierarchical matrix which is either random graph or constant block
    in the following way:

    First sample groups according to the stick breaking construction (like a DP
    except just giving each group an index). Make sure that there are at least
    two groups. The DP parameter is alpha (i.e. this is proportional to the
    probability of forming a new group). Assign all intercluster similarities to
    the the first value in p (in the case of constant block) or bernoulli(p) in
    the case of random graph. Recursively partition each cluster in the same
    way.

    Parameters:
    n = number of points in the matrix
    p = either a list or a single number representing probabilities (r = True)
    at each level in the hierarchy, or probabilities at the outermost level
    which then get increased as we go down the hierarchy. Or they represent
    intercluster similarities (starting low and increasing).
    alpha = dirichlet parameter (probability of creating new group)
    r = boolean, if True then make random graph, otherwise constant block.
    """

    if n == 1:
        return (CountingMatrix(np.matrix(np.ones((1,1)))), cluster_tree.ClusterTree([[0]]), "ahb")
    groups = util.dirichlet_groups(n, alpha)
    while (max(groups) == 0):
        groups = util.dirichlet_groups(n, alpha)
    M = np.matrix(np.zeros((n,n)))
    if type(p) == list:
        currp = p[0]
        if len(p) == 1:
            nextp = p[0]
        else:
            nextp = p[1:len(p)]
    else:
        currp = p
        nextp = p+float(1-p)/2
    for i in range(n):
        for j in range(i,n):
            if groups[i] != groups[j]:
                if r == False:
                    M[i,j] = currp
                    M[j,i] = currp
                else:
                    M[i,j] = np.random.binomial(1, currp)
                    M[j,i] = M[i,j]
    clustertree = []
    for i in range(max(groups)+1):
        inds = [x for x in range(len(groups)) if groups[x] == i]
        if len(inds) == 0:
            continue
        (Msub, clusters, blah) = make_arbitrary_hierarchical_block(len(inds), nextp, alpha, r)
        clusters = clusters.lists
        M[np.ix_(inds, inds)] = Msub.data
        clustertree.append(clusters)
    clustertree.reverse()
    clustertree.append(groups)
    clustertree.reverse()                
    return (CountingMatrix(M), cluster_tree.ClusterTree(clustertree), "ahb")

def make_overlapping(n, p, q, overlap, noisy=True):
    """
    This is an overlapping planted partition model. Effectively it is the
    planted partition, except "overlap" number of nodes are in both clusters
    (i.e. they have probability p of edges to either clusters).
    """
    if noisy:
        M = np.matrix(np.random.binomial(1, q, [n,n]))
        M = np.triu(M,k=1) + np.triu(M,k=1).T 
        A = np.matrix(np.random.binomial(1, p, [(n+overlap)/2, (n+overlap)/2]))
        A = np.triu(A,k=1) + np.triu(A,k=1).T
        M[np.ix_(range(0,(n+overlap)/2), range(0,(n+overlap)/2))] = A
        A = np.matrix(np.random.binomial(1, p, [(n+overlap)/2, (n+overlap)/2]))
        A = np.triu(A,k=1) + np.triu(A,k=1).T
        M[np.ix_(range((n-overlap)/2,n), range((n-overlap)/2,n))] = A
        M += np.matrix(np.diagflat([1 for i in range(n)]))
    else:
        M = np.matrix(q*np.ones([n,n]))
        A = np.matrix(p*np.ones([(n+overlap)/2, (n+overlap)/2]))
        M[np.ix_(range(0,(n+overlap)/2), range(0,(n+overlap)/2))] = A        
        M[np.ix_(range((n-overlap)/2,n), range((n-overlap)/2,n))] = A
    return CountingMatrix(M)
    
def make_gaussian_noise(n, sigma):
    """
    This makes a symmetric gaussian noise matrix with mean 0 and variance sigma
    """
    R = np.matrix(np.random.normal(0, sigma, (n,n)))
    R = 1.0/2*(R+R.T)
    return R
